# handles interactions with JSON files
import json
from . import crypto_utils
from datetime import date
import os

# admin + user DB JSON initialization
def init_admin_DB(admin_username, admin_pwrd):
    crypto_text = crypto_utils.hash_admin_creds(admin_pwrd)     
    DB ={ "username": admin_username,
         "ciphertext": crypto_text
        }
    
    with open(f"{admin_username}.json", "w") as f:
        json.dump(DB, f)
    print(f"New admin: {admin_username} successfully added. Vault initialized and ready for use.")
def init_user_DB(admin_username, admin_pwrd): 
    empty_entries_DB = {"entries": []}
    entries_json = json.dumps(empty_entries_DB)
    
    dat = date.today()
    year = dat.year
    
    DB_salt = crypto_utils.generate_salt()
    DB_key = crypto_utils.derive_key(admin_pwrd, DB_salt)

    entries = crypto_utils.encrypt(DB_key, entries_json)
    DB_key = None
    
    DB= {"vault_ID": f"{admin_username}_vault_{year}",
         "kdf": {"algo": "argon2_id",
                 "salt_base64": DB_salt.decode("utf-8"),
                 "time": 3,
                 "memory": 65536,
                 "parallelism": 4,
                 "hash_length": 32},
        "contents": entries.decode()
    }
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB,f)

# user <--> DB interactions
def add_creds(admin_username, admin_pwrd, new_entries: dict): 
    # decrypting the database
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    entries = DB_header["contents"]
    kdf = DB_header["kdf"]
    salt = kdf["salt_base64"]
    salt = salt.encode("utf-8")
    
    key = crypto_utils.derive_key(admin_pwrd, salt)
    entries_plaintext = crypto_utils.decrypt(key, entries)
    entries_plaintext = entries_plaintext.decode()
    entries_plaintext= json.loads(entries_plaintext)
    
    # adding the new credentials to the DB
    entries_list = entries_plaintext["entries"]
    entries_list.append(new_entries)
    
    # re-encrypting the user DB entries
    entries_plaintext = json.dumps(entries_plaintext)
    entries_token = crypto_utils.encrypt(key, entries_plaintext)
    key = None
    DB_header["contents"] = entries_token.decode()
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)
    print("New credentials successfully added!")

def delete_creds(admin_username, admin_pwrd, entry_ID):
   # decrypting the database
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    entries = DB_header["contents"]
    kdf = DB_header["kdf"]
    salt = kdf["salt_base64"]
    salt = salt.encode("utf-8")
    
    key = crypto_utils.derive_key(admin_pwrd, salt)
    entries_plaintext = crypto_utils.decrypt(key, entries)
    entries_plaintext = entries_plaintext.decode()
    entries_plaintext= json.loads(entries_plaintext)
    
    # deleting an entry based on the ID(corresponds to list index)
    entries_list = entries_plaintext["entries"]
    entries_list.pop(entry_ID)
    
    # re-encrypting the user DB entries
    entries_json = json.dumps(entries_plaintext)
    entries_token = crypto_utils.encrypt(key, entries_json)
    key = None
    DB_header["contents"] = entries_token.decode()
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)
    print(f"The credentials of entry {entry_ID} were successfully deleted.")

def edit_creds(admin_username, admin_pwrd, field, newinfo, entry_ID):
    # decrypting the database
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    entries = DB_header["contents"]
    kdf = DB_header["kdf"]
    salt = kdf["salt_base64"]
    salt = salt.encode("utf-8")
    
    DBkey = crypto_utils.derive_key(admin_pwrd, salt)
    entries_plaintext = crypto_utils.decrypt(DBkey, entries)
    entries_plaintext = entries_plaintext.decode()
    entries_plaintext= json.loads(entries_plaintext) 

    # updating the specified credentials with new info
    entries_list = entries_plaintext["entries"]
    entry = entries_list[entry_ID] # note: entry is a dictionary
    
    if field == "password":
            salt = crypto_utils.generate_salt()
            key = crypto_utils.derive_key(admin_pwrd, salt)
            token = crypto_utils.encrypt(key, newinfo)
            key = None
            
            entry["password"] = token.decode()
            entry["salt"] = salt.decode()
    else:
        entry[field] = newinfo
    
    # re-encrypting the user DB entries
    entries_json = json.dumps(entries_plaintext)
    entries_token = crypto_utils.encrypt(DBkey, entries_json)
    DBkey = None
    DB_header["contents"] = entries_token.decode()
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)
    print(f"The {field} of entry {entry_ID} was successfully updated.")

def list_creds(admin_username, admin_pwrd):
    # decrypting the database
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    entries = DB_header["contents"]
    kdf = DB_header["kdf"]
    salt = kdf["salt_base64"]
    salt = salt.encode("utf-8")
    
    key = crypto_utils.derive_key(admin_pwrd, salt)
    entries_plaintext = crypto_utils.decrypt(key, entries)
    entries_plaintext = entries_plaintext.decode()
    entries_plaintext= json.loads(entries_plaintext)
    
    # printing the creds from the database
    entries_list = entries_plaintext["entries"]
    if len(entries_list) ==0:
        print("The vault is empty.")
    else:
        print("Credentials:")
        for i in entries_list:
            print(f"""\nEntry_ID: {entries_list.index(i)}
Website: {i["website"]}
Username: {i["username"]}
Password: {"*"*7}""") # arbitrary value
    
    # re-encrypting the user DB entries
    entries_json = json.dumps(entries_plaintext)
    entries_token = crypto_utils.encrypt(key, entries_json)
    key = None
    DB_header["contents"] = entries_token.decode()
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)

def view_pwrd(admin_username, admin_pwrd, entry_ID):
    # decrypting the database
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    entries = DB_header["contents"]
    kdf = DB_header["kdf"]
    DBsalt = kdf["salt_base64"]
    DBsalt = DBsalt.encode("utf-8") 
    
    DBkey = crypto_utils.derive_key(admin_pwrd, DBsalt)
    entries_plaintext = crypto_utils.decrypt(DBkey, entries)
    entries_plaintext = entries_plaintext.decode()
    entries_plaintext= json.loads(entries_plaintext)
    
    # accessing specified entry
    entries_list = entries_plaintext["entries"]
    entry = entries_list[entry_ID]
    
    # decrypting the password
    password_token = entry["password"]
    salt = entry["salt"]
    salt = salt.encode("utf-8")
    key = crypto_utils.derive_key(admin_pwrd, salt)
    password = crypto_utils.decrypt(key, password_token)
    key = None
    
    # printing the information
    print(f"""\nEntry_ID: {entry_ID}
Website: {entry["website"]}
Username: {entry["username"]}
Password: {password.decode()}""") 
    
    password = None #bye-bye reference :)
    
    # re-encrypting the user DB entries
    entries_json = json.dumps(entries_plaintext)
    entries_token = crypto_utils.encrypt(DBkey, entries_json)
    DBkey = None
    DB_header["contents"] = entries_token.decode()
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)

def reset_DB(admin_username, admin_pwrd):
    # decrypting the database
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    entries = DB_header["contents"]
    kdf = DB_header["kdf"]
    salt = kdf["salt_base64"]
    salt = salt.encode("utf-8")
    
    key = crypto_utils.derive_key(admin_pwrd, salt)
    entries_plaintext = crypto_utils.decrypt(key, entries)
    entries_plaintext = entries_plaintext.decode()
    
    # overwriting the database as empty
    entries_plaintext= {"entries":[]}
    
    # re-encrypting the user DB entries
    entries_json = json.dumps(entries_plaintext)
    entries_token = crypto_utils.encrypt(key, entries_json)
    key = None
    DB_header["contents"] = entries_token.decode()
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)
    print(f"{admin_username}'s vault has been reset.")

def list_by_site(admin_username, admin_pwrd, site):
    # decrypting the database
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    entries = DB_header["contents"]
    kdf = DB_header["kdf"]
    salt = kdf["salt_base64"]
    salt = salt.encode("utf-8")
    
    key = crypto_utils.derive_key(admin_pwrd, salt)
    entries_plaintext = crypto_utils.decrypt(key, entries)
    entries_plaintext = entries_plaintext.decode()
    entries_plaintext= json.loads(entries_plaintext)
    
    #printing credentials based on desired website
    entries_list = entries_plaintext["entries"]
    site_list = []
    for i in entries_list:
        if i["website"] == site:
            site_list.append(i)
    if len(site_list)==0:
        print(f"There are no saved credentials for {site}")
    else:
        for i in site_list:
                print(f"""\nEntry_ID: {entries_list.index(i)}
    Website: {i["website"]}
    Username: {i["username"]}
    Password: {"*"*7}""")
        
    # re-encrypting the user DB entries
    entries_json = json.dumps(entries_plaintext)
    entries_token = crypto_utils.encrypt(key, entries_json)
    key = None
    DB_header["contents"] = entries_token.decode()
    
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)
    
def delete_admin(admin_username):
    # sanitizing the user database before deletion
    with open(f"{admin_username}_vault.json", "r") as f:
        DB_header = json.load(f)
    
    DB_header["contents"] = None
    kdf = DB_header["kdf"]
    kdf["salt_base64"] = None
 
    with open(f"{admin_username}_vault.json", "w") as f:
        json.dump(DB_header,f)
       
    # sanitizing the admin file before deletion
    with open(f"{admin_username}.json", "r") as f:
        admin_json = json.load(f)
    
    admin_json["ciphertext"] = None
 
    with open(f"{admin_username}.json", "w") as f:
        json.dump(admin_json,f)
        
    # deleting the files
    os.remove(f"{admin_username}_vault.json")
    os.remove(f"{admin_username}.json")
    print(f"{admin_username}'s vault files have been successfully deleted.")


# handles database and plaintext encryption, decryption, salt generation, all things encryption + hashing
import argon2
import base64
from cryptography.fernet import Fernet
import os

# admin specific functions
def hash_admin_creds(admin_password):
    PH = argon2.PasswordHasher()
    hash = PH.hash(admin_password)
    return hash

def verify_admin(admin_hash, input_password):
    PH = argon2.PasswordHasher()
    return PH.verify(admin_hash, input_password)

# encryption dependencies - these return byte-like object for en/decryption use, or to be decoded and stored in desired location
def generate_salt ():
    salt = os.urandom(16)
    base_64_salt = base64.b64encode(salt)
    return base_64_salt 

def derive_key(admin_pwrd, salt):
    secret = admin_pwrd.encode("utf-8")
    key = argon2.low_level.hash_secret_raw(secret, salt, 3, 65536, 4, 32, argon2.low_level.Type.ID)
    base_64_key = base64.urlsafe_b64encode(key)
    return base_64_key

# database en/decryption fuctions - all en/decryption functions take decoded bytes, so they have to encode them before encryption
def encrypt(key, plain):
    plain_bytes = plain.encode("utf-8")
    f = Fernet(key)
    ciphertoken = f.encrypt(plain_bytes)
    return ciphertoken

def decrypt(key, cipher_token):
    cipher_token.encode("utf-8")
    f = Fernet(key)
    plain = f.decrypt(cipher_token)
    return plain



# handles the packaging + passing of inputs to core functions
from . import auth_manager
from . import crypto_utils
from . import storage

# admin specific functions
def add_admin(admin_username, admin_pwrd):
    admin = storage.init_admin_DB(admin_username, admin_pwrd)
    user = storage.init_user_DB(admin_username, admin_pwrd)
    return admin, user

# database functions
def add_creds(admin_username, admin_password, new_username, new_website, new_password):
    if auth_manager.verify_admin(admin_username, admin_password):
        #encrypting the password and packaging all the new credentials into dict
        print("Adding new entry ...")
        salt = crypto_utils.generate_salt()
        key = crypto_utils.derive_key(admin_password, salt)
        token = crypto_utils.encrypt(key, new_password)
        
        new_creds = {"username": new_username,
       "website": new_website,
       "salt": salt.decode(),
       "password": token.decode()}

        return storage.add_creds(admin_username, admin_password, new_creds)

def delete_creds(admin_username, admin_password, entry_ID):
    if auth_manager.verify_admin(admin_username, admin_password): 
        print(f"Deleting entry {entry_ID}...") 
        return storage.delete_creds(admin_username, admin_password, int(entry_ID))

def update_creds(admin_username, admin_password, entry_ID, field, new_info ):
    if auth_manager.verify_admin(admin_username, admin_password): 
        print(f"Updating entry {entry_ID}...")  
        return storage.edit_creds(admin_username, admin_password, field, new_info, int(entry_ID))

def list_all(admin_username, admin_password):
    if auth_manager.verify_admin(admin_username, admin_password): 
        return storage.list_creds(admin_username, admin_password) 
    
def list_by_site(admin_username, admin_password, site):
    if auth_manager.verify_admin(admin_username, admin_password): 
        return storage.list_by_site(admin_username, admin_password, site)
    
def see_pwrd(admin_username, admin_password, entry_ID):
    if auth_manager.verify_admin(admin_username, admin_password): 
        return storage.view_pwrd(admin_username, admin_password, int(entry_ID))
  
def reset_DB(admin_username, admin_password):
    if auth_manager.verify_admin(admin_username, admin_password): 
        print(f"Resetting {admin_username}'s vault...") 
        return storage.reset_DB(admin_username, admin_password)

def delete_admin(admin_username, admin_password):
    if auth_manager.verify_admin(admin_username, admin_password): 
        # second confirmation conditional
        confirm2 = input("Please understand that you are NOT reseting the vault.\nYou are DELETING your vault and admin account.\Do you wish to proceed? Y/N? ")
        if confirm2 == "Y" or confirm2 == "y":
            print(f"Deleting admin account{admin_username}...") 
            return storage.delete_admin(admin_username)
        else: 
            print("Delete admin request cancelled.")





# handles authentication passing
from . import crypto_utils
import json

def verify_admin(input_username, input_pwrd) -> bool: 
        with open(f"{input_username}.json") as DB:
            credentials = json.load(DB)
        valid = crypto_utils.verify_admin(credentials["ciphertext"], input_pwrd)
        return valid





# handles argument parsing + collecting inputs
import argon2
import argparse
from getpass import getpass
from . import input_manager
import os

def main():
    # all parsers
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest = "command")

    add_admin_parser = subparsers.add_parser("newadmin", help = "add a new admin user(creates new vault)")
    add_admin_parser.add_argument("--username", required=True)
    add_admin_parser.add_argument("--password")

    add_creds_parser = subparsers.add_parser("newcreds", help = "add new credentials for an application/website")
    add_creds_parser.add_argument("--site", required=True)
    add_creds_parser.add_argument("--username", required=True)
    add_creds_parser.add_argument("--password")

    del_parser = subparsers.add_parser("del", help ="delete a selected user credential")
    del_parser.add_argument("--entryID", required =True)

    update_parser = subparsers.add_parser("update", help = "update existing credentials for fields: password username site")
    update_parser.add_argument("--entryID", required=True)
    update_parser.add_argument("--field", required=True)
    update_parser.add_argument("--newinfo", required=True)

    list_parser = subparsers.add_parser("list", help="list all of the existing credentials")

    list_site_parser = subparsers.add_parser("sitelist", help ="list credentials belonging to a specified site")
    list_site_parser.add_argument("--site", required=True)

    view_parser = subparsers.add_parser("view", help = "view a specific password")
    view_parser.add_argument("--entryID", required = True)

    reset_parser = subparsers.add_parser("reset", help="WARNING: this will RESET the vault, deleting all credentials in the vault.")

    del_admin_parser = subparsers.add_parser("deladmin", help="WARNING: this will DELETE the admin account. Vault will no longer exist.")


    args = parser.parse_args()

    try:
        if args.command == "newadmin":
            if os.path.exists(f"{args.username}_vault.json") or os.path.exists(f"{args.username}.json"):
                confirm_overwrite = input(f"There already exists a vault under admin name: \n{args.username} \nDo you want to overwrite it? Y/N? ")
                if confirm_overwrite == "Y" or confirm_overwrite == "y":
                    if args.password is None:
                        args.password = getpass(f"Enter a password for {args.username}: ")
                    print("Creating new vault ...")
                    input_manager.add_admin(args.username, args.password)
                else:
                    print("New admin request cancelled. Consider choosing a different admin username.")
            else:
                if args.password is None:
                    args.password = getpass(f"Enter a password for {args.username}: ")
                print("Creating new vault ...")
                input_manager.add_admin(args.username, args.password)

            
        elif args.command == "newcreds":
            if args.password is None:
                args.password = getpass(f"Enter a password for the new entry: ")

            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
            input_manager.add_creds(admin_username, admin_password, args.username, args.site, args.password)
            
        elif args.command == "del":
            # confirm and execute or cancel delete action
            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
                
            confirm = input(f"Are you sure you want to delete entry {args.entryID}?\nThis action cannot be undone. Y/N? ")
            if confirm == "Y" or confirm == "y":
                input_manager.delete_creds(admin_username, admin_password, args.entryID)
            else:
                print("Delete request cancelled.")
            
        elif args.command == "update":
            # sanitize inputs
            whitelist = ["password", "username", "site"]
            while args.field not in whitelist:
                args.field = input("You've entered an invalid field.\nValid fields are: \npassword\nusername\nsite\nPlease enter a valid field: ")
            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
            input_manager.update_creds(admin_username, admin_password, args.entryID, args.field, args.newinfo)
            
        elif args.command == "list":
            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
            input_manager.list_all(admin_username, admin_password)
            
        elif args.command == "sitelist":
            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
            input_manager.list_by_site(admin_username, admin_password, args.site)
            
        elif args.command == "view":
            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
            input_manager.see_pwrd(admin_username, admin_password, args.entryID)
            
        elif args.command == "reset":
            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
            confirm = input(f"Are you sure you want to reset the vault for Admin {admin_username}?\nThis action cannot be undone. Y/N? ")
            if confirm == "Y" or confirm == "y":
                input_manager.reset_DB(admin_username, admin_password)
            else:
                print("Reset request cancelled.")

        elif args.command == "deladmin":
            # first confirm and execute or cancel delete admin action
            admin_username = input("Please enter the admin username: ")
            admin_password = getpass("Please enter the admin password: ")
                
            confirm1 = input(f"Are you sure you want to delete {admin_username}'s vault?\nThis action cannot be undone. Y/N? ")
            if confirm1 == "Y" or confirm1 == "y":
                input_manager.delete_admin(admin_username, admin_password)
            else:
                print("Delete admin request cancelled.")        
        else:
            parser.print_help()
    except argon2.exceptions.VerifyMismatchError:
        print("Incorrect password entered.")
    except FileNotFoundError:
        print("The vault for the admin username does not exist.")
    except ValueError:
        print("Please be sure to enter integer values for entryID.")

if __name__ == "__main__":
    main()


    
